from __future__ import annotations


def expend_range(current_range, new_range):
    if current_range is None:
        return new_range

    return [
        min(current_range[0], new_range[0]),
        max(current_range[1], new_range[1]),
    ]



from __future__ import annotations

from pathlib import Path

BASE_DIRECTORY = Path(__file__).parent.parent.parent.parent
INPUT = "data/model_0000000878.hdf5"

SAMPLE_DATASET = (BASE_DIRECTORY / INPUT).resolve()



from __future__ import annotations

import argparse
from pathlib import Path

from vtkmodules.vtkCommonCore import vtkPoints
from vtkmodules.vtkCommonDataModel import (
    vtkCellArray,
    vtkPolyData,
)
from vtkmodules.vtkIOXML import vtkXMLPolyDataWriter

from parsli.utils import earth


def to_polydata(input_file: Path, use_spherical_proj):
    vtk_points = vtkPoints()
    vtk_lines = vtkCellArray()
    dataset = vtkPolyData(points=vtk_points, lines=vtk_lines)
    insert = earth.insert_spherical if use_spherical_proj else earth.insert_euclidian

    with input_file.open() as txt_file:
        line = txt_file.readline()
        while line:
            if "segment" in line:
                n_points = int(line.split("points")[1])
                vtk_lines.InsertNextCell(n_points)
                for _ in range(n_points):
                    line = txt_file.readline()
                    lat, lon = map(float, line.split())
                    vtk_lines.InsertCellPoint(insert(vtk_points, lon, lat, 0))
            else:
                print("Don't know what to do with line:", line)  # noqa: T201

            line = txt_file.readline()

    return dataset


def write_file(input_file, use_spherical_proj):
    projection = "spherical" if use_spherical_proj else "euclidean"
    output_name = str(input_file.parent / f"{input_file.stem}-{projection}.vtp")

    writer = vtkXMLPolyDataWriter()
    writer.file_name = output_name
    writer.input_data = to_polydata(input_file, use_spherical_proj)
    writer.SetCompressorTypeToZLib()
    writer.SetDataModeToBinary()
    writer.SetCompressionLevel(6)
    writer.Write()


def main():
    parser = argparse.ArgumentParser(
        prog="parsli.tools.wdb",
        description="Parsli conversion tool to generate vtp files from coastline text file",
    )
    (
        parser.add_argument(
            "--input", help="Path to the wdb file to convert", required=True
        ),
    )
    parser.add_argument(
        "-s", "--spherical", help="Project lat/lon on sphere", action="store_true"
    )
    args, _ = parser.parse_known_args()
    input_file = Path(args.input).resolve()
    if not input_file.exists():
        print(f"Invalid input path: {input_file}")  # noqa: T201
        return

    use_spherical_proj = args.spherical
    if input_file.is_dir:
        for f in input_file.glob("*.txt"):
            write_file(f, use_spherical_proj)
    elif input_file.is_file:
        write_file(input_file, use_spherical_proj)
    else:
        print("Invalid input type")  # noqa: T201


if __name__ == "__main__":
    main()



from __future__ import annotations

from trame.decorators import TrameApp, change
from trame.widgets import html
from trame.widgets import vuetify3 as v3
from vtkmodules.vtkCommonDataModel import vtkDataObject, vtkDataSetAttributes

from parsli.utils import expend_range
from parsli.viewer.vtk import PRESETS, set_preset, to_image


@TrameApp()
class ControlPanel(v3.VCard):
    def __init__(self, toggle, scene_manager, reset_camera, reset_to_mesh):
        self._scene_manager = scene_manager

        super().__init__(
            classes="controller",
            elevation=5,
            rounded=(f"{toggle} || 'circle'",),
        )

        # allocate variable if does not exist
        self.state.setdefault(toggle, True)

        with self:
            with v3.VCardTitle(
                classes=(
                    f"`d-flex align-center pa-1 position-fixed bg-white ${{ {toggle} ? 'controller-content rounded-t border-b-thin':'rounded-circle'}}`",
                ),
                style="z-index: 1;",
            ):
                v3.VProgressLinear(
                    v_if=toggle,
                    indeterminate=("trame__busy",),
                    bg_color="rgba(0,0,0,0)",
                    absolute=True,
                    color="primary",
                    location="bottom",
                    height=2,
                )
                v3.VProgressCircular(
                    v_else=True,
                    bg_color="rgba(0,0,0,0)",
                    indeterminate=("trame__busy",),
                    style="position: absolute; top: 0; left: 0; width: 100%; height: 100%;",
                    color="primary",
                    width=3,
                )
                v3.VBtn(
                    icon="mdi-close",
                    v_if=toggle,
                    click=f"{toggle} = !{toggle}",
                    flat=True,
                    size="sm",
                )
                v3.VBtn(
                    icon="mdi-menu",
                    v_else=True,
                    click=f"{toggle} = !{toggle}",
                    # flat=True,
                    variant="outlined",
                    size="sm",
                )

                html.Div(
                    "Control Panel",
                    v_show=toggle,
                    classes="text-h6 px-2",
                )
                v3.VSpacer()
                v3.VCheckbox(
                    v_show=toggle,
                    v_model=("show_segment", True),
                    true_icon="mdi-gesture",
                    false_icon="mdi-gesture",
                    hide_details=True,
                    density="compact",
                )
                v3.VCheckbox(
                    v_show=toggle,
                    v_model=("show_surface", True),
                    true_icon="mdi-texture-box",
                    false_icon="mdi-texture-box",
                    hide_details=True,
                    density="compact",
                    classes="mx-2",
                )

            with v3.VCardText(
                v_show=(toggle, True),
                classes="controller-content mt-12 mb-1 pb-1 mx-0 px-1",
            ):
                # -------------------------------------------------------------
                # Longitude / Latitude cropping
                # -------------------------------------------------------------

                with v3.VRow(
                    "Longitude", classes="text-subtitle-2 ma-1 pt-2 align-center"
                ):
                    v3.VSpacer()
                    html.Span(
                        "{{ longitude_bnds[0].toFixed(1) }}",
                        classes="text-caption text-center",
                        style="width: 2.5rem;",
                    )
                    html.Span(
                        "{{ longitude_bnds[1].toFixed(1) }}",
                        classes="text-caption text-center",
                        style="width: 2.5rem;",
                    )

                v3.VRangeSlider(
                    v_model=("longitude_bnds", [0, 360]),
                    min=0,
                    max=360,
                    step=0.5,
                    density="compact",
                    hide_details=True,
                    classes="px-2",
                )

                with v3.VRow("Latitude", classes="text-subtitle-2 ma-1 align-center"):
                    v3.VSpacer()
                    html.Span(
                        "{{ latitude_bnds[0].toFixed(1) }}",
                        classes="text-caption text-center",
                        style="width: 2.5rem;",
                    )
                    html.Span(
                        "{{ latitude_bnds[1].toFixed(1) }}",
                        classes="text-caption text-center",
                        style="width: 2.5rem;",
                    )

                v3.VRangeSlider(
                    v_model=("latitude_bnds", [-90, 90]),
                    min=-90,
                    max=90,
                    step=0.5,
                    density="compact",
                    hide_details=True,
                    classes="px-2",
                )

                with v3.VRow(classes="ma-1"):
                    v3.VBtn(
                        icon="mdi-crop-free",
                        size="small",
                        flat=True,
                        density="compact",
                        hide_details=True,
                        click=reset_camera,
                    )
                    v3.VBtn(
                        icon="mdi-magnify-scan",
                        size="small",
                        flat=True,
                        density="compact",
                        hide_details=True,
                        classes="mx-2",
                        click=reset_to_mesh,
                    )

                    v3.VSpacer()
                    v3.VBtn(
                        icon="mdi-map-plus",
                        size="small",
                        flat=True,
                        density="compact",
                        hide_details=True,
                        classes="mx-2",
                        click=self._expand_bounds,
                    )

                    v3.VBtn(
                        icon="mdi-arrow-collapse-horizontal",
                        size="small",
                        flat=True,
                        density="compact",
                        hide_details=True,
                        click=self._crop_bounds_to_mesh,
                    )

                    v3.VBtn(
                        icon="mdi-arrow-expand-horizontal",
                        size="small",
                        flat=True,
                        density="compact",
                        hide_details=True,
                        classes="mx-2",
                        click=self._reset_bounds,
                    )

                # -------------------------------------------------------------

                v3.VDivider(classes="mt-2 mx-n3")

                # -------------------------------------------------------------
                # Projection: Spherical / Euclidean
                # -------------------------------------------------------------

                with html.Div(classes="d-flex"):
                    v3.VSelect(
                        prepend_inner_icon=(
                            "spherical ? 'mdi-earth' : 'mdi-earth-box'",
                        ),
                        v_model=("spherical", True),
                        items=(
                            "proj_modes",
                            [
                                {"title": "Spherical", "value": True},
                                {"title": "Euclidean", "value": False},
                            ],
                        ),
                        hide_details=True,
                        density="compact",
                        flat=True,
                        variant="solo",
                        classes="mx-n2",
                    )
                    v3.VCheckbox(
                        disabled=("!spherical",),
                        v_model=("show_earth_core", True),
                        true_icon="mdi-google-earth",
                        false_icon="mdi-google-earth",
                        hide_details=True,
                        density="compact",
                    )

                # -------------------------------------------------------------
                # Coast line regions
                # -------------------------------------------------------------

                v3.VSelect(
                    prepend_inner_icon="mdi-map-outline",
                    placeholder="Coast lines",
                    v_model=("coast_active_regions", []),
                    items=("coast_regions", []),
                    density="compact",
                    hide_details=True,
                    flat=True,
                    variant="solo",
                    chips=True,
                    closable_chips=True,
                    multiple=True,
                    classes="mx-n2",
                )
                # -------------------------------------------------------------

                v3.VDivider(classes="mx-n3 mb-1")

                # -------------------------------------------------------------
                # Color mapping
                # -------------------------------------------------------------

                v3.VSelect(
                    placeholder="Color By",
                    prepend_inner_icon="mdi-format-color-fill",
                    v_model=("color_by", "dip_slip"),
                    items=("fields", []),
                    hide_details=True,
                    density="compact",
                    flat=True,
                    variant="solo",
                    classes="mx-n2",
                )

                with v3.VRow(no_gutters=True, classes="align-center mx-n2 mt-n2"):
                    with v3.VCol():
                        v3.VTextField(
                            v_model_number=("color_min", 0),
                            type="number",
                            hide_details=True,
                            density="compact",
                            flat=True,
                            variant="solo",
                            hide_spin_buttons=True,
                        )
                    with html.Div(classes="flex-0 mx-n3", style="z-index: 1;"):
                        v3.VBtn(
                            icon="mdi-arrow-expand-horizontal",
                            size="sm",
                            density="compact",
                            flat=True,
                            variant="solo",
                            classes="ml-2",
                            click=self.reset_color_range,
                        )
                        v3.VBtn(
                            icon="mdi-circle-half-full",
                            size="sm",
                            density="compact",
                            flat=True,
                            variant="solo",
                            classes="mr-2",
                            click=self.symetric_color_range,
                        )
                    with v3.VCol():
                        v3.VTextField(
                            v_model_number=("color_max", 1),
                            type="number",
                            hide_details=True,
                            density="compact",
                            flat=True,
                            variant="solo",
                            reverse=True,
                            classes="px-0",
                            hide_spin_buttons=True,
                        )

                html.Img(
                    src=("preset_img", None),
                    style="height: 1rem; width: 100%;",
                    classes="rounded-lg border-thin mb-n1",
                )

                v3.VSelect(
                    placeholder="Color Preset",
                    prepend_inner_icon="mdi-palette",
                    v_model=("color_preset", "Rycroft"),
                    items=("color_presets", list(PRESETS.keys())),
                    hide_details=True,
                    density="compact",
                    flat=True,
                    variant="solo",
                    classes="mx-n2",
                )

                v3.VDivider(classes="mx-n3 pb-1")

                # -------------------------------------------------------------
                # Contours
                # -------------------------------------------------------------

                with v3.VTooltip(
                    text=("`Number of contours: ${nb_contours}`",),
                ):
                    with html.Template(v_slot_activator="{ props }"):
                        with html.Div(
                            classes="d-flex pr-2",
                            v_bind="props",
                        ):
                            v3.VSlider(
                                v_model=("nb_contours", 5),
                                min=2,
                                max=20,
                                step=1,
                                prepend_icon="mdi-fingerprint",
                                hide_details=True,
                                density="compact",
                                flat=True,
                                variant="solo",
                                style="margin: 0 2px;",
                            )

    def _reset_bounds(self):
        self.state.latitude_bnds = [-90, 90]
        self.state.longitude_bnds = [0, 360]
        self.ctrl.view_update()

    def _expand_bounds(self):
        delta = 0.5
        self.state.latitude_bnds = [
            max(self.state.latitude_bnds[0] - delta, -90),
            min(self.state.latitude_bnds[1] + delta, 90),
        ]
        delta /= 2
        self.state.longitude_bnds = [
            max(self.state.longitude_bnds[0] - delta, 0),
            min(self.state.longitude_bnds[1] + delta, 360),
        ]
        self.ctrl.view_update()

    def _crop_bounds_to_mesh(self):
        source = self._scene_manager["meshes"].get("source")
        self.state.latitude_bnds = source.latitude_bounds
        self.state.longitude_bnds = source.longitude_bounds
        self.ctrl.view_update()

    @change("coast_active_regions")
    def _on_regions(self, coast_active_regions, **_):
        source = self._scene_manager["coast"].get("source")
        source.active_regions = coast_active_regions
        self.ctrl.view_update()

    @change("show_segment", "show_surface", "show_earth_core")
    def _on_visibility(self, show_segment, show_surface, show_earth_core, **_):
        seg_actors = self._scene_manager["segment"].get("actors")
        surf_actors = self._scene_manager["meshes"].get("actors")
        earth_actors = self._scene_manager["earth_core"].get("actors")

        for actor in seg_actors:
            actor.SetVisibility(show_segment)

        for actor in surf_actors:
            actor.SetVisibility(show_surface)

        for actor in earth_actors:
            actor.SetVisibility(show_earth_core)

        self.ctrl.view_update()

    @change("latitude_bnds", "longitude_bnds")
    def _on_lat_lon_bnd(self, longitude_bnds, latitude_bnds, **_):
        if self._scene_manager["segment"]:
            reader = self._scene_manager["segment"].get("source")
            reader.longitude_bnds = longitude_bnds
            reader.latitude_bnds = latitude_bnds
        self.ctrl.view_update()

    @change("color_by", "color_preset", "color_min", "color_max", "nb_contours")
    def _on_color_preset(
        self, color_preset, color_min, color_max, color_by, nb_contours, **_
    ):
        lut = self._scene_manager.lut
        color_min = float(color_min)
        color_max = float(color_max)

        for mesh_type in ["segment", "meshes"]:
            if self._scene_manager[mesh_type]:
                mapper = self._scene_manager[mesh_type].get("mapper")
                mapper.SelectColorArray(color_by)
                mapper.SetScalarVisibility(1)
                mapper.SetScalarRange(color_min, color_max)

        mesh_pipeline = self._scene_manager["meshes"]
        assign = mesh_pipeline.get("assign")
        assign.Assign(
            color_by,
            vtkDataSetAttributes.SCALARS,
            vtkDataObject.FIELD_ASSOCIATION_POINTS,
        )
        bands = mesh_pipeline.get("bands")
        bands.GenerateValues(nb_contours, [color_min, color_max])

        if "color_preset" in self.state.modified_keys:
            set_preset(lut, color_preset)
            self.state.preset_img = to_image(lut, 255)

        self.ctrl.view_update()

    def symetric_color_range(self):
        max_bound = max(abs(self.state.color_min), abs(self.state.color_max))
        self.state.color_min = -max_bound
        self.state.color_max = max_bound

    def reset_color_range(self):
        pipeline_item = self._scene_manager["meshes"]
        source = pipeline_item.get("source")
        ds = source()

        total_range = None
        for array in ds.cell_data[self.state.color_by].Arrays:
            total_range = expend_range(total_range, array.GetRange())

        self.state.color_min = total_range[0]
        self.state.color_max = total_range[1]


class ViewToolbar(v3.VCard):
    def __init__(self, reset_camera, reset_to_mesh, apply_zoom, update_view_up):  # noqa: ARG002
        super().__init__(
            classes="view-toolbar pa-1",
            rounded="lg",
        )

        with self:
            with v3.VTooltip(text="Reset camera"):
                with html.Template(v_slot_activator="{ props }"):
                    v3.VBtn(
                        v_bind="props",
                        flat=True,
                        density="compact",
                        icon="mdi-crop-free",
                        click=reset_camera,
                    )

            with v3.VTooltip(text="Reset camera centered on mesh"):
                with html.Template(v_slot_activator="{ props }"):
                    v3.VBtn(
                        v_bind="props",
                        flat=True,
                        density="compact",
                        icon="mdi-magnify-scan",
                        click=reset_to_mesh,
                    )
            v3.VDivider()
            with v3.VTooltip(text="Zoom in"):
                with html.Template(v_slot_activator="{ props }"):
                    v3.VBtn(
                        v_bind="props",
                        flat=True,
                        density="compact",
                        icon="mdi-magnify-plus-outline",
                        click=(apply_zoom, "[1.1]"),
                    )
            with v3.VTooltip(text="Zoom out"):
                with html.Template(v_slot_activator="{ props }"):
                    v3.VBtn(
                        v_bind="props",
                        flat=True,
                        density="compact",
                        icon="mdi-magnify-minus-outline",
                        click=(apply_zoom, "[0.9]"),
                    )

            # --------------------------------------------
            # Not needed if we use terrain style
            # --------------------------------------------
            # v3.VDivider()
            # with v3.VTooltip(text="Z up"):
            #     with html.Template(v_slot_activator="{ props }"):
            #         v3.VBtn(
            #             v_bind="props",
            #             flat=True,
            #             density="compact",
            #             icon="mdi-axis-z-arrow",
            #             click=(update_view_up, "[[0, 0, 1]]"),
            #         )
            # --------------------------------------------



from __future__ import annotations

import math

EARTH_RADIUS = 6371.0


def insert_spherical(vtk_points, longitude, latitude, depth):
    longitude = math.pi * longitude / 180
    latitude = math.pi * latitude / 180
    h = EARTH_RADIUS - depth
    return vtk_points.InsertNextPoint(
        h * math.cos(longitude) * math.cos(latitude),
        h * math.sin(longitude) * math.cos(latitude),
        h * math.sin(latitude),
    )


def insert_euclidian(vtk_point, longitude, latitude, depth):
    return vtk_point.InsertNextPoint(
        longitude,
        latitude,
        depth,
    )


def interpolate(a_longitude, a_latitude, b_longitude, b_latitude, distance):
    a_xyz = (
        EARTH_RADIUS * math.cos(a_longitude) * math.cos(a_latitude),
        EARTH_RADIUS * math.sin(a_longitude) * math.cos(a_latitude),
        EARTH_RADIUS * math.sin(a_latitude),
    )
    b_xyz = (
        EARTH_RADIUS * math.cos(b_longitude) * math.cos(b_latitude),
        EARTH_RADIUS * math.sin(b_longitude) * math.cos(b_latitude),
        EARTH_RADIUS * math.sin(b_latitude),
    )
    ab_xyz = (b_xyz[0] - a_xyz[0], b_xyz[1] - a_xyz[1], b_xyz[2] - a_xyz[2])
    ab_dist = math.sqrt(
        ab_xyz[0] * ab_xyz[0] + ab_xyz[1] * ab_xyz[1] + ab_xyz[2] * ab_xyz[2]
    )
    ratio = distance / ab_dist
    return (
        ratio * (b_longitude - a_longitude) + a_longitude,
        ratio * (b_latitude - a_latitude) + a_latitude,
    )



from __future__ import annotations

import logging
from pathlib import Path

import h5py
from vtkmodules.util.vtkAlgorithm import VTKPythonAlgorithmBase
from vtkmodules.vtkCommonCore import vtkPoints
from vtkmodules.vtkCommonDataModel import (
    vtkCellArray,
    vtkPartitionedDataSet,
    vtkPolyData,
)

from parsli.utils import earth

logger = logging.getLogger(__name__)
# logger.setLevel(logging.DEBUG)


class VtkMeshReader(VTKPythonAlgorithmBase):
    def __init__(self):
        VTKPythonAlgorithmBase.__init__(
            self,
            nInputPorts=0,
            nOutputPorts=1,
            outputType="vtkPartitionedDataSet",
        )
        self._file_name = None
        self._proj_spherical = True
        self._longitude_bnd = [0, 360]
        self._latitude_bnd = [-90, 90]

    @property
    def file_name(self):
        return self._file_name

    @file_name.setter
    def file_name(self, path):
        self._file_name = Path(path)
        if not self._file_name.exists():
            msg = f"Invalid file path: {self._file_name.resolve()}"
            raise ValueError(msg)

        self.Modified()

    @property
    def spherical(self):
        return self._proj_spherical

    @spherical.setter
    def spherical(self, value):
        if self._proj_spherical != value:
            self._proj_spherical = value
            self.Modified()

    @property
    def available_fields(self):
        if self._file_name is None or not self._file_name.exists():
            return []

        result = set()
        with h5py.File(self._file_name, "r") as hdf:
            meshes = hdf["meshes"]
            for mesh in meshes:
                for name in meshes[mesh]:
                    if name in {"mesh_name", "coordinates", "verts"}:
                        continue
                    result.add(name)

        return list(result)

    @property
    def longitude_bounds(self):
        return self._longitude_bnd

    @property
    def latitude_bounds(self):
        return self._latitude_bnd

    def _expend_bounds(self, longitude, latitude):
        self._longitude_bnd[0] = min(longitude, self._longitude_bnd[0])
        self._longitude_bnd[1] = max(longitude, self._longitude_bnd[1])

        self._latitude_bnd[0] = min(latitude, self._latitude_bnd[0])
        self._latitude_bnd[1] = max(latitude, self._latitude_bnd[1])

    def RequestData(self, _request, _inInfo, outInfo):
        if self._file_name is None or not self._file_name.exists():
            return 1

        # Reset bounds
        self._longitude_bnd = [360, 0]
        self._latitude_bnd = [90, -90]

        # Read file and generate mesh
        output = self.GetOutputData(outInfo, 0)
        all_meshes = vtkPartitionedDataSet()

        # Projection selection
        insert_pt = earth.insert_spherical if self.spherical else earth.insert_euclidian

        with h5py.File(self._file_name, "r") as hdf:
            meshes = hdf["meshes"]
            n_meshes = len(meshes)
            logger.debug("n_meshes=%s", n_meshes)
            all_meshes.SetNumberOfPartitions(n_meshes)
            for idx, mesh in enumerate(meshes):
                vtk_mesh = vtkPolyData()
                all_meshes.SetPartition(idx, vtk_mesh)

                # Process known names first
                # - coordinates
                vtk_points = vtkPoints()
                vtk_points.SetDataTypeToDouble()
                vtk_mesh.points = vtk_points

                hdf_ds = meshes[mesh]["coordinates"]
                n_points = hdf_ds.shape[0]
                vtk_points.Allocate(n_points)

                for xyz in hdf_ds:
                    self._expend_bounds(xyz[0], xyz[1])
                    insert_pt(vtk_points, xyz[0], xyz[1], -xyz[2])

                # - verts
                vtk_polys = vtkCellArray()
                vtk_mesh.polys = vtk_polys

                hdf_ds = meshes[mesh]["verts"]
                n_cells = hdf_ds.shape[0]
                vtk_polys.Allocate(4 * n_cells)
                assert hdf_ds.shape[1] == 3, "only triangles"

                logger.debug(" %s. mesh %s cells", idx, n_cells)

                for cell in hdf_ds:
                    vtk_polys.InsertNextCell(3)
                    vtk_polys.InsertCellPoint(cell[0])
                    vtk_polys.InsertCellPoint(cell[1])
                    vtk_polys.InsertCellPoint(cell[2])

                for name in meshes[mesh]:
                    if name in {"mesh_name", "coordinates", "verts"}:
                        continue

                    # This is a field
                    logger.debug("Field %s: %s", name, meshes[mesh][name].shape)
                    vtk_mesh.cell_data[name] = meshes[mesh][name][:].ravel()

        output.ShallowCopy(all_meshes)
        return 1



from __future__ import annotations

from pathlib import Path

from trame.app import get_server
from trame.decorators import TrameApp, change
from trame.ui.vuetify3 import VAppLayout
from trame.widgets import vtk as vtkw
from trame.widgets import vtklocal
from trame.widgets import vuetify3 as v3
from vtkmodules.vtkFiltersSources import vtkSphereSource
from vtkmodules.vtkIOParallelXML import vtkXMLPartitionedDataSetWriter

from parsli.io import VtkCoastLineSource, VtkMeshReader, VtkSegmentReader
from parsli.utils import expend_range
from parsli.utils.earth import EARTH_RADIUS
from parsli.viewer import css, ui
from parsli.viewer.vtk import SceneManager

DEBUG_WRITE_MESH = False


@TrameApp()
class Viewer:
    def __init__(self, server=None):
        self.server = get_server(server, client_type="vue3")
        self.server.enable_module(css)
        self.server.cli.add_argument(
            "--data", help="Path of hdf5 file to load", required=True
        )
        self.server.cli.add_argument(
            "--wasm", help="Use local rendering", action="store_true"
        )

        # process cli
        args, _ = self.server.cli.parse_known_args()
        data_file = str(Path(args.data).resolve())
        self.local_rendering = args.wasm

        # Setup app
        self.scene_manager = SceneManager(self.server)
        self._build_ui()

        # earth core
        pipeline = self.scene_manager.add_geometry(
            "earth_core",
            vtkSphereSource(
                radius=EARTH_RADIUS - 100,
                theta_resolution=60,
                phi_resolution=60,
            ),
        )
        prop = pipeline.get("actor").property
        prop.opacity = 0.85

        # load segments
        seg_reader = VtkSegmentReader()
        seg_reader.file_name = data_file
        pipeline = self.scene_manager.add_geometry("segment", seg_reader)
        pipeline.get("mapper").SetScalarModeToUseCellFieldData()

        # load meshes
        mesh_reader = VtkMeshReader()
        mesh_reader.file_name = data_file
        pipeline = self.scene_manager.add_geometry_with_contour(
            "meshes", mesh_reader, True
        )
        pipeline.get("mapper").SetScalarModeToUsePointFieldData()
        self.state.fields = mesh_reader.available_fields

        # Coast lines
        self.coast_lines = VtkCoastLineSource()
        self.state.coast_regions = self.coast_lines.available_regions
        self.state.coast_active_regions = []
        pipeline = self.scene_manager.add_geometry("coast", self.coast_lines, True)
        coast_props = pipeline.get("actor").property
        coast_props.line_width = 2
        coast_props.color = (0, 0, 0)

        if DEBUG_WRITE_MESH:
            writer = vtkXMLPartitionedDataSetWriter()
            writer.SetInputData(mesh_reader())
            writer.SetFileName("all_meshes.vtpd")
            writer.Write()

        self.readers = [mesh_reader, seg_reader]

    @property
    def ctrl(self):
        return self.server.controller

    @property
    def state(self):
        return self.server.state

    @change("color_by")
    def _on_color_by(self, color_by, **_):
        pipeline_item = self.scene_manager["meshes"]
        source = pipeline_item.get("source")
        mapper_mesh = pipeline_item.get("mapper")
        mapper_seg = self.scene_manager["segment"].get("mapper")

        if color_by is None:
            mapper_mesh.SetScalarVisibility(0)
            mapper_seg.SetScalarVisibility(0)
            self.ctrl.view_update()
            return

        # Extract data range
        ds = source()

        total_range = None
        for array in ds.cell_data[color_by].Arrays:
            total_range = expend_range(total_range, array.GetRange())

        self.state.color_min = total_range[0]
        self.state.color_max = total_range[1]

    @change("spherical")
    def _on_projection_change(self, spherical, **_):
        self.state.show_earth_core = spherical

        for geo_name in ["segment", "meshes", "coast"]:
            pipeline_item = self.scene_manager[geo_name]
            pipeline_item.get("source").spherical = spherical
            actors = pipeline_item.get("actors")

            scale = (1, 1, 1) if spherical else (1, 1, 0.01)
            for actor in actors:
                actor.scale = scale

        self.ctrl.view_reset_camera()

    @change("camera")
    def _on_camera(self, camera, **_):
        if camera is None:
            return

        self.ctrl.vtk_update_from_state(camera)

    def reset_to_mesh(self):
        bounds = self.scene_manager["meshes"].get("actor").bounds
        self.scene_manager.reset_camera_to(bounds)
        self.ctrl.view_update(push_camera=True)

    def apply_zoom(self, scale):
        self.scene_manager.apply_zoom(scale)
        self.ctrl.view_update(push_camera=True)

    def update_view_up(self, view_up):
        self.scene_manager.update_view_up(view_up)
        self.ctrl.view_update(push_camera=True)

    def _build_ui(self):
        self.state.trame__title = "Parsli"
        self.state.setdefault("camera", None)
        with VAppLayout(self.server, full_height=True) as layout:
            self.ui = layout

            with v3.VContainer(
                fluid=True, classes="fill-height pa-0 ma-0 position-relative"
            ):
                if self.local_rendering:
                    with vtklocal.LocalView(
                        self.scene_manager.render_window,
                        20,
                        camera="camera = $event",
                    ) as view:
                        view.register_vtk_object(self.scene_manager.widget)
                        self.ctrl.view_update = view.update
                        self.ctrl.view_reset_camera = view.reset_camera
                        self.ctrl.vtk_update_from_state = view.vtk_update_from_state
                else:
                    with vtkw.VtkRemoteView(
                        self.scene_manager.render_window,
                        interactive_ratio=2,
                        still_ratio=2,
                    ) as view:
                        self.ctrl.view_update = view.update
                        self.ctrl.view_reset_camera = view.reset_camera

                # Control panel
                ui.ControlPanel(
                    toggle="show_panel",
                    scene_manager=self.scene_manager,
                    reset_camera=self.ctrl.view_reset_camera,
                    reset_to_mesh=self.reset_to_mesh,
                )

                # 3D View controls
                ui.ViewToolbar(
                    reset_camera=self.ctrl.view_reset_camera,
                    reset_to_mesh=self.reset_to_mesh,
                    apply_zoom=self.apply_zoom,
                    update_view_up=self.update_view_up,
                )
